package org.processmining.glueminer.algorithms;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.processmining.glueminer.models.Behaviors;
import org.processmining.glueminer.models.OrderedActivity;
import org.processmining.glueminer.models.PlaceUsageStatus;
import org.processmining.glueminer.models.PlacesMemorizer;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

public class PlaceAdditionTriager {
	Behaviors behaviors;
	Petrinet net;
	Marking marking;
	Marking finalMarking;
	PlacesMemorizer placesMemorizer;
	Map<String, Transition> activitiesTransition;
	
	Integer noOfPlacesAdded;
	Integer noOfPlacesRemoved;
	Integer noOfPlacesKept;
	
	public PlaceAdditionTriager(Behaviors behaviors, Petrinet net, Marking initialMarking, Marking finalMarking, PlacesMemorizer placesMemorizer, Map<String, Transition> activitiesTransition) {
		this.behaviors = behaviors;
		this.net = net;
		this.marking = new Marking(initialMarking);
		this.finalMarking = finalMarking;
		this.placesMemorizer = placesMemorizer;
		this.activitiesTransition = activitiesTransition;
		
		this.noOfPlacesAdded = 0;
		this.noOfPlacesRemoved = 0;
		this.noOfPlacesKept = 0;
	}
	
	public Set<String> getSetStringFromSetOrderedActivity(Set<OrderedActivity> activities0) {
		Set<String> activities = new HashSet<String>();
		
		for (OrderedActivity activity : activities0) {
			activities.add(activity.activity);
		}
		
		return activities;
	}
	
	public Set<Transition> getSetTransitionFromSetString(Set<String> activities0) {
		Set<Transition> activities = new HashSet<Transition>();
		
		for (String activity : activities0) {
			activities.add(activitiesTransition.get(activity));
		}
		
		return activities;
	}
	
	public boolean checkRedundant(Set<OrderedActivity> inputActivities, Set<OrderedActivity> outputActivities) {
		return placesMemorizer.checkRedundancy(getSetStringFromSetOrderedActivity(inputActivities), getSetStringFromSetOrderedActivity(outputActivities));
	}
	
	public boolean checkOtherPlacesFitness(TokenReplayer replayer, Double minimumPlaceFitness, Double maximumPlaceUnderfedness) {
		for (Place place : net.getPlaces()) {
			if (place.getLabel() != "source" && place.getLabel() != "sink") {
				if (replayer.placeFitness.get(place) < minimumPlaceFitness) {
					return false;
				}
				else if (replayer.placeUnderfedness.get(place) > maximumPlaceUnderfedness) {
					return false;
				}
			}
		}
		
		return true;
	}
	
	public Object[] tryAddPlace(Set<OrderedActivity> inputActivities0, Set<OrderedActivity> outputActivities0, Double minimumPlaceFitness, Double maximumPlaceUnderfedness) {
		this.noOfPlacesAdded++;
		
		Set<String> inputActivities = getSetStringFromSetOrderedActivity(inputActivities0);
		Set<String> outputActivities = getSetStringFromSetOrderedActivity(outputActivities0);
		
		String placeLabel = placesMemorizer.getPlaceLabel(inputActivities, outputActivities);
		
		Set<Transition> inputTransitions = getSetTransitionFromSetString(inputActivities);
		Set<Transition> outputTransitions = getSetTransitionFromSetString(outputActivities);
		
		Place currentlyEvaluatedPlace = net.addPlace(placeLabel);
		List<Arc> arcsToPlace = addArcsToPlace(inputTransitions, outputTransitions, currentlyEvaluatedPlace);
		
		TokenReplayer replayer = new TokenReplayer(this.behaviors, this.net, this.marking, this.finalMarking);
		replayer.doReplay();
		PlaceUsageStatus.status status = replayer.placeUsageStatus.get(currentlyEvaluatedPlace);
		double placeFitness = replayer.placeFitness.get(currentlyEvaluatedPlace);
		double placeUnderfedness = replayer.placeUnderfedness.get(currentlyEvaluatedPlace);
				
		if (placeFitness < minimumPlaceFitness) {
			this.noOfPlacesRemoved++;
			removeElements(currentlyEvaluatedPlace, arcsToPlace);
		}
		else if (placeUnderfedness > maximumPlaceUnderfedness) {
			this.noOfPlacesRemoved++;
			removeElements(currentlyEvaluatedPlace, arcsToPlace);
		}
		else {
			if (checkOtherPlacesFitness(replayer, minimumPlaceFitness, maximumPlaceUnderfedness)) {
				this.noOfPlacesKept++;
				placesMemorizer.addPlace(inputActivities, outputActivities);
			}
			else {
				this.noOfPlacesRemoved++;
				removeElements(currentlyEvaluatedPlace, arcsToPlace);
			}
		}
		
		return new Object[] { status, placeFitness, placeUnderfedness };
	}
	
	public void removeElements(Place place, List<Arc> arcsToPlace) {
		for (Arc arc : arcsToPlace) {
			net.removeEdge(arc);
		}
		net.removePlace(place);
	}
	
	public List<Arc> addArcsToPlace(Set<Transition> inputTransitions, Set<Transition> outputTransitions, Place place) {
		List<Arc> arcsToPlace = new ArrayList<Arc>();
		
		for (Transition input : inputTransitions) {
			Arc arc = (Arc) net.addArc(input, place);
			arcsToPlace.add(arc);
		}
		for (Transition output : outputTransitions) {
			Arc arc = (Arc) net.addArc(place, output);
			arcsToPlace.add(arc);
		}
		return arcsToPlace;
	}

}
